package com.wombat.aliyun_oss.service

import android.content.Context
import android.net.Uri
import android.util.Log
import com.alibaba.sdk.android.oss.ClientException
import com.alibaba.sdk.android.oss.OSS
import com.alibaba.sdk.android.oss.ServiceException
import com.alibaba.sdk.android.oss.callback.OSSCompletedCallback
import com.alibaba.sdk.android.oss.callback.OSSProgressCallback
import com.alibaba.sdk.android.oss.common.OSSHeaders
import com.alibaba.sdk.android.oss.common.OSSLog
import com.alibaba.sdk.android.oss.model.OSSRequest
import com.alibaba.sdk.android.oss.model.ObjectMetadata
import com.alibaba.sdk.android.oss.model.ResumableDownloadRequest
import com.alibaba.sdk.android.oss.model.ResumableDownloadResult
import com.alibaba.sdk.android.oss.model.ResumableUploadRequest
import com.alibaba.sdk.android.oss.model.ResumableUploadResult
import com.wombat.aliyun_oss.listener.*
import com.wombat.aliyun_oss.model.download.*
import com.wombat.aliyun_oss.model.upload.*
import com.wombat.aliyun_oss.model.upload.UploadSuccessData
import com.wombat.aliyun_oss.utils.RealPathUtil
import java.io.File
import java.net.URLEncoder


class OSSService private constructor(
    private val oss: OSS,
    private val recordDirectory: String,
    private val uploadCallback: UploadCallback,
    private val downloadCallback: DownloadCallback,
    private val context: Context
) {

    private var uploadTasks = mutableSetOf<UploadTaskData>()
    private var downloadTasks = mutableSetOf<DownloadTaskData>()

    companion object {
        @Volatile
        private var INSTANCE: OSSService? = null

        fun getInstance(
            oss: OSS,
            recordDirectory: String,
            uploadCallback: UploadCallback,
            downloadCallback: DownloadCallback,
            context: Context
        ): OSSService {
            if (INSTANCE == null) {
                synchronized(this) {
                    if (INSTANCE == null) {
                        INSTANCE =
                            OSSService(
                                oss,
                                recordDirectory,
                                uploadCallback,
                                downloadCallback,
                                context
                            )
                    }
                }
            }
            return INSTANCE!!
        }
    }

    fun upload(
        bucketName: String,
        objectKey: String,
        fileUri: Uri,
        uuid: String,
        mimeType: String?
    ) {
        val uploadStart = System.currentTimeMillis()
        OSSLog.logDebug("upload start")
        if (objectKey.isEmpty()) {
            Log.w("AsyncUploadRequest", "ObjectNull")
            return
        }
        // 构造上传请求
        OSSLog.logDebug("create ResumableUploadRequest ")

        val recordDir = File(recordDirectory)

        // 确保断点记录的保存文件夹已存在，如果不存在则新建断点记录的保存文件夹。
        if (!recordDir.exists()) {
            recordDir.mkdirs()
        }

        // 添加文件名 + 文件md5 + 文件类型 + 文件大小
        val metadata = ObjectMetadata().apply {
            // 文件名
            val fileName = RealPathUtil.getFileName(context, fileUri)
            val encodeFileName =
                URLEncoder.encode(fileName ?: "", "utf-8")
            Log.d("文件原名", "$fileName")
            addUserMetadata(OSSHeaders.OSS_USER_METADATA_PREFIX + "file-name", encodeFileName)
            // 内容md5, 断点续传不支持md5
//            contentMD5 = BinaryUtil.base64FromHexString(objectKey.split(".")[0])
            // 文件类型
            if (mimeType != null) {
                Log.d("文件类型", "$mimeType")
                contentType = mimeType
            }
        }

        val rq =
            ResumableUploadRequest(
                bucketName,
                objectKey,
                fileUri,
                metadata,
                recordDirectory
            )

        rq.setDeleteUploadOnCancelling(true)
        rq.crC64 = OSSRequest.CRC64Config.YES

        rq.progressCallback =
            OSSProgressCallback<ResumableUploadRequest> { request, currentSize, totalSize ->
                Log.d("进度数据", "${request.metadata.rawMetadata}\n${request.metadata.userMetadata}")
                val uploadProgress = UploadProgressData(
                    objectKey = objectKey,
                    uuid = uuid,
                    uploadId = request.uploadId,
                    totalBytesSent = currentSize,
                    totalBytesExpectedToSend = totalSize
                )
                uploadCallback.progressCallback(uploadProgress)
            }


        val task = oss.asyncResumableUpload(
            rq,
            object :
                OSSCompletedCallback<ResumableUploadRequest, ResumableUploadResult> {
                override fun onSuccess(
                    request: ResumableUploadRequest,
                    result: ResumableUploadResult
                ) {
                    val uploadEnd = System.currentTimeMillis()
                    Log.d("上传耗时", "${(uploadEnd - uploadStart) / 1000f}")
                    Log.d("成功结果", "${request.metadata} \n${result.responseHeader}")

                    val success = UploadSuccessData(
                        objectKey = objectKey,
                        uuid = uuid,
                        uploadId = request.uploadId,
                        result = result.location
                    )
                    uploadCallback.successCallback(success)

                    uploadTasks.filter {
                        it.objectKey == objectKey
                    }.forEach {
                        uploadTasks.remove(it)
                    }

                }

                override fun onFailure(
                    request: ResumableUploadRequest,
                    clientException: ClientException?,
                    serviceException: ServiceException?
                ) {
                    var info = ""
                    // 请求异常
                    if (clientException != null) {
                        // 本地异常如网络异常等
                        clientException.printStackTrace()
                        info = clientException.toString()
                    }
                    if (serviceException != null) {
                        // 服务异常
                        info = serviceException.toString()
                    }
                    uploadCallback.failCallback(
                        UploadFailData(
                            objectKey = objectKey,
                            uuid = uuid,
                            uploadId = request.uploadId,
                            error = info
                        )
                    )
                }
            })

        uploadTasks.add(
            UploadTaskData(
                objectKey,
                uuid,
                rq.uploadId,
                bucketName,
                fileUri,
                task,
                mimeType
            )
        )

    }

    fun pauseUpload(
        objectKey: String,
        uuid: String
    ): String? {
        return uploadTasks.firstOrNull {
            it.uuid == uuid
        }?.run {
            task.cancel()
            objectKey
        }

    }

    fun resumeUpload(
        objectKey: String,
        uuid: String
    ): String? {
        return uploadTasks.firstOrNull {
            it.uuid == uuid
        }?.run {
            uploadTasks.remove(this)
            upload(bucketName, objectKey, fileUri, uuid, mimeType)
            bucketName
        }
    }

    fun cancelUpload(objectKey: String, uuid: String): String? {
        return uploadTasks.firstOrNull {
            it.uuid == uuid
        }?.run {
            uploadTasks.remove(this)
            this.task.cancel()
            bucketName
        }

    }

    fun cancelAllUpload(): String? {
        uploadTasks.forEach {
            it.task.cancel()
        }
        uploadTasks.clear()
        return null
    }

    fun download(
        bucketName: String,
        objectKey: String,
        fileUri: Uri
    ) {
        val downloadStart = System.currentTimeMillis()
        Log.d("开始下载", "$objectKey")
        if (objectKey.isEmpty()) {
            Log.w("AsyncDownloadRequest", "ObjectNull")
            return
        }

        val rq =
            ResumableDownloadRequest(
                bucketName,
                objectKey,
                fileUri.path,
                recordDirectory
            )

        rq.progressListener =
            OSSProgressCallback<ResumableDownloadRequest> { request, currentSize, totalSize ->
                val downloadProgress = DownloadProgressData(
                    objectKey = objectKey,
                    bucketName = request.bucketName,
                    totalBytesSent = currentSize,
                    totalBytesExpectedToSend = totalSize
                )
                downloadCallback.progressCallback(downloadProgress)
            }

        val task = oss.asyncResumableDownload(
            rq,
            object :
                OSSCompletedCallback<ResumableDownloadRequest, ResumableDownloadResult> {
                override fun onSuccess(
                    request: ResumableDownloadRequest,
                    result: ResumableDownloadResult
                ) {
                    val downloadEnd = System.currentTimeMillis()
                    Log.d("下载耗时", "${(downloadEnd - downloadStart) / 1000f}")

                    val success = DownloadSuccessData(
                        objectKey = objectKey,
                        bucketName = request.bucketName,
                        result = result.metadata
                    )
                    downloadCallback.successCallback(success)

                    downloadTasks.filter {
                        it.objectKey == objectKey
                    }.forEach {
                        downloadTasks.remove(it)
                    }

                }

                override fun onFailure(
                    request: ResumableDownloadRequest,
                    clientException: ClientException?,
                    serviceException: ServiceException?
                ) {
                    var info = ""
                    // 请求异常
                    if (clientException != null) {
                        // 本地异常如网络异常等
                        clientException.printStackTrace()
                        info = clientException.toString()
                    }
                    if (serviceException != null) {
                        // 服务异常
                        info = serviceException.toString()
                    }
                    downloadCallback.failCallback(
                        DownloadFailData(
                            objectKey = objectKey,
                            bucketName = request.bucketName,
                            error = info
                        )
                    )
                }
            })

        downloadTasks.add(DownloadTaskData(objectKey, bucketName, fileUri, task))

    }

    fun pauseDownload(
        objectKey: String
    ): String? {
        return downloadTasks.firstOrNull {
            it.objectKey == objectKey
        }?.run {
            task.cancel()
            objectKey
        }
    }

    fun resumeDownload(
        objectKey: String
    ): String? {
        return downloadTasks.firstOrNull {
            it.objectKey == objectKey
        }?.run {
            downloadTasks.remove(this)
            download(bucketName, objectKey, fileUri)
            bucketName
        }


    }
}